#include <vector>
#include <algorithm>
#include <set>
#include <map>
#include <utility>
#include <iterator>
#include "common.h"
using namespace std;

class kratka
{
	int n_mozliwosci;
	vector<bool> mozliwosci;
	int pierwsza_mozliwosc;
	bool ok;

	void init(int i);
public:
	kratka(int r = 0) : n_mozliwosci(r), mozliwosci(r, true), pierwsza_mozliwosc(0), ok(false) {}
	kratka& operator=(int i) { init(i); return *this; }

	bool rozwiazana() const { return ok; }
	int liczba_mozliwosci() const { return n_mozliwosci; }
	int wartosc() const { return pierwsza_mozliwosc; }
	bool mozliwa(int v) const { return mozliwosci[v]; }
	bool usun_mozliwosc(int v);
	bool usun_mozliwosci(const kratka& b);
	bool dodaj_mozliwosc(int v);
	void dodaj_mozliwosci(const kratka& b);
	static bool eq(const kratka& a, const kratka& b);

	bool operator==(const kratka& b) const { return eq(*this, b); }
	bool operator!=(const kratka& b) const { return !eq(*this, b); }
	kratka& operator|=(const kratka& b) { dodaj_mozliwosci(b); return *this; }

	class iterator_mozliwosci;
	friend class kratka::iterator_mozliwosci;
	class iterator_mozliwosci
	{
		const vector<bool>& mozliwosci;
		int i, n;
	public:
		iterator_mozliwosci(const kratka& k)
			: mozliwosci(k.mozliwosci), i(k.pierwsza_mozliwosc), n(k.n_mozliwosci)
		{
			if (n < 1)
				throw int(0);
		}

		operator bool() const { return n > 0; }
		int operator*() const { return i; }
		iterator_mozliwosci& operator++() { if (--n > 0) while(!mozliwosci[++i]); return *this; }
	};

	void dump(ostream& s, const zadanie& z) const;
};

typedef set<int> zbior_kratek;

void kratka::init(int i)
{
	if (!n_mozliwosci) {
		mozliwosci.resize(n_mozliwosci = i);
		fill(mozliwosci.begin(), mozliwosci.end(), true);
	} else {
		fill(mozliwosci.begin(), mozliwosci.end(), false);
		n_mozliwosci = 1;
		ok = true;
		mozliwosci[pierwsza_mozliwosc = i] = true;
	}
}

bool kratka::usun_mozliwosc(int v)
{
	if (!mozliwosci[v])
		return false;

	--n_mozliwosci;
	mozliwosci[v] = false;

	if (!n_mozliwosci)
		throw int(0);

	while (!mozliwosci[pierwsza_mozliwosc])
		++pierwsza_mozliwosc;

	return n_mozliwosci == 1;
}

bool kratka::usun_mozliwosci(const kratka& b)
{
	for(kratka::iterator_mozliwosci p(b); p; ++p) {
		if (!mozliwosci[*p])
			continue;

		--n_mozliwosci;
		mozliwosci[*p] = false;
	}

	if (!n_mozliwosci)
		return false;

	while (!mozliwosci[pierwsza_mozliwosc])
		++pierwsza_mozliwosc;

	return true;
}

bool kratka::dodaj_mozliwosc(int v)
{
	if (mozliwosci[v])
		return false;

	++n_mozliwosci;
	mozliwosci[v] = true;

	if (pierwsza_mozliwosc > v)
		pierwsza_mozliwosc = v;

	return true;
}

void kratka::dodaj_mozliwosci(const kratka& b)
{
	for(kratka::iterator_mozliwosci p(b); p; ++p) {
		if (!mozliwosci[*p]) {
			mozliwosci[*p] = true;
			++n_mozliwosci;
		}
	}

	if (pierwsza_mozliwosc > b.pierwsza_mozliwosc)
		pierwsza_mozliwosc = b.pierwsza_mozliwosc;
}

bool kratka::eq(const kratka& a, const kratka& b)
{
	if (a.n_mozliwosci != b.n_mozliwosci)
		return false;
	
	for(kratka::iterator_mozliwosci pa(a), pb(b); pa; ++pa, ++pb) {
		if (*pa != *pb)
			return false;
	}

	return true;
}

#ifdef DEBUG
void kratka::dump(ostream& ds, const zadanie& z) const
{
	ds << '[';
	for(unsigned i = 0; i < mozliwosci.size(); ++i) {
		ds << (mozliwosci[i] ? z.chars[i] : ' ');
		if (ok && mozliwosci[i])
			ds << '\b' << z.chars[i];
	}
	ds << ']';
}

static void dump_full_field(const zadanie& z, const vector<kratka>& pole)
{
	vector<kratka>::const_iterator p = pole.begin();

	for(int y = 0; y < z.rr; ++y) {
		for(int x = 0; x < z.rr; ++x) {
			const kratka& k = *(p++);

			k.dump(cerr, z); cerr << ' ';
			if (((x+1) % z.rzad) == 0)
				cerr << "  ";
		}
		cerr << endl;
		if (((y+1) % z.rzad) == 0)
			cerr << endl;
	}
}

static void dump_pos(const zadanie& z, int nr)
{
	cerr << ((nr % z.rr)+1) << ',' << ((nr / z.rr)+1);
}

static void dump_zk(const zadanie& z, const zbior_kratek& kz)
{
	for(zbior_kratek::const_iterator dp = kz.begin(); dp != kz.end(); ++dp) {
		cerr << ' ';
		dump_pos(z, *dp);
	}
}

#endif /* DEBUG */

class kombinator
{
	zadanie& z;
	int pozostalo;
	vector<kratka> pole;

	vector< zbior_kratek > zaleznosci;	/* grupy zależnych kratek */
	map< int, set<int> > zkratki;		/* grupy, do których należy kratka */
	set< pair<int, int> > pary_zaleznosci;	/* pary grup, które mają > 1 kratkę wspólną */

	void dodaj_mozliwosc(int nr, int zi);
	void wygeneruj_zaleznosci();
	void wpisz_w_kratke(int nr, int v, zbior_kratek& do_wpisania);
	void wpisz_poczatkowe_kratki(zbior_kratek& do_wpisania);
	void wpisz_odgadniete(zbior_kratek& do_wpisania);
	void optymalizuj_kratki(zbior_kratek& do_wpisania);
	void optymalizuj_grupe(zbior_kratek& grupa, zbior_kratek& do_wpisania);
	bool optymalizuj_podgrupe(zbior_kratek& kz, vector<int>& znaki, zbior_kratek& do_wpisania);
	void optymalizuj_wspolne(zbior_kratek& g1, zbior_kratek& g2, kratka wspolne, zbior_kratek& do_wpisania);
	bool optymalizuj_miedzy_grupami(zbior_kratek& grupa1, zbior_kratek& grupa2, zbior_kratek& do_wpisania);
public:
	kombinator(zadanie& _z)
		: z(_z), pozostalo(z.rr * z.rr), pole(pozostalo, kratka(z.rr))
	{}

	bool kombinuj();
	void zapisz_rozwiazanie(rozwiazanie& r);
};

void kombinator::dodaj_mozliwosc(int nr, int zi)
{
	if (pole[nr].rozwiazana())
		return;

	zaleznosci[zi].insert(nr);
	zkratki[nr].insert(zi);
}

template <typename T, typename C = unsigned int>
class counting_iterator
{
	C &counter_value;
public:
	counting_iterator(C &counter) : counter_value(counter) {}
	counting_iterator& operator++() { ++counter_value; return *this; }
	counting_iterator& operator++(int) { ++counter_value; return *this; }
	counting_iterator& operator*() { return *this; }
	const T& operator=(const T& x) const { return x; }
	T& operator=(T& x) { return x; }
};

void kombinator::wygeneruj_zaleznosci()
{
	int nn = z.rr * 3;	/* wiersze,kolumny,kwadraty */
	if (z.add_diag)
		nn += 2;	/* przekątne */
	zaleznosci.resize(nn);
	int zi = 0;

	/* wiersze */
	for(int y = 0, nr = 0; y < z.rr; ++y, ++zi) {
		for(int x = 0; x < z.rr; ++x, ++nr)
			dodaj_mozliwosc(nr, zi);
	}

	/* kolumny */
	for(int x = 0; x < z.rr; ++x, ++zi) {
		for(int y = 0, nr = x; y < z.rr; ++y, (nr += z.rr))
			dodaj_mozliwosc(nr, zi);
	}

	/* kwadraty */
	for(int dy = 0; dy < z.rzad; ++dy) {
		for(int dx = 0; dx < z.rzad; ++dx, ++zi) {
			int nr = (dy * z.rr + dx ) * z.rzad;
			for(int y = 0; y < z.rzad; ++y, (nr += z.rr - z.rzad)) {
				for(int x = 0; x < z.rzad; ++x, ++nr)
					dodaj_mozliwosc(nr, zi);
			}
		}
	}

	/* przekatne */
	if (z.add_diag) {
		for(int i = 0, nr = 0; i < z.rr; ++i, (nr += z.rr + 1))
			dodaj_mozliwosc(nr, zi);
		++zi;
		for(int i = 0, nr = z.rr - 1; i < z.rr; ++i, (nr += z.rr - 1))
			dodaj_mozliwosc(nr, zi);
		++zi;
	}

	/* pary */
	pair< int, int > p;
	for(p.first = 0; p.first < zi; ++p.first) {
		zbior_kratek &pf = zaleznosci[p.first];

		for(p.second = p.first + 1; p.second < zi; ++p.second) {
			zbior_kratek &ps = zaleznosci[p.second];

			unsigned common = 0;
			set_intersection(pf.begin(), pf.end(), ps.begin(), ps.end(),
				counting_iterator<zbior_kratek::value_type>(common));
			if (common > 1)
				pary_zaleznosci.insert(p);
		}
	}
}

void kombinator::wpisz_w_kratke(int nr, int v, zbior_kratek& do_wpisania)
{
	kratka& k = pole[nr];

	if (k.rozwiazana()) {
		if (v == k.wartosc())
			return;
		throw int(0);
	}

	/* wpisz */
#ifdef DEBUG
	cerr << "SET: "; dump_pos(z, nr); cerr << " -> " << z.chars[v] << ' '; k.dump(cerr, z); cerr << endl;
#endif
	k = v;

	/* znajdź wszystkie zależne kratki (i usuń tę z grupy) */
	zbior_kratek zalezne;
	set<int>& grupy = zkratki[nr];
	for(set<int>::iterator p = grupy.begin(); p != grupy.end(); ++p) {
		zbior_kratek& kz = zaleznosci[*p];
		kz.erase(nr);
		zalezne.insert(kz.begin(), kz.end());
	}
	grupy.clear();

	for(zbior_kratek::iterator p = zalezne.begin(); p != zalezne.end(); ++p) {
#ifdef DEBUG
		try {
#endif
			if (pole[*p].usun_mozliwosc(v))
				do_wpisania.insert(*p);
#ifdef DEBUG
		} catch (int) {
			cerr << "ZEROed: "; dump_pos(z, *p); cerr << " -= " << z.chars[v] << endl;
			throw;
		}
#endif
	}

	--pozostalo;
}

void kombinator::wpisz_poczatkowe_kratki(zbior_kratek& do_wpisania)
{
	for(vector<init_pt>::const_iterator p = z.init.begin(); p != z.init.end(); ++p)
		wpisz_w_kratke(p->nr, p->v, do_wpisania);
}

void kombinator::wpisz_odgadniete(zbior_kratek& do_wpisania)
{
	while(!do_wpisania.empty()) {
		int nr = *do_wpisania.begin();
		do_wpisania.erase(do_wpisania.begin());

		wpisz_w_kratke(nr, pole[nr].wartosc(), do_wpisania);
	}
}

void kombinator::optymalizuj_kratki(zbior_kratek& do_wpisania)
{
	for(vector<zbior_kratek>::iterator p = zaleznosci.begin(); p != zaleznosci.end(); ++p) {
		zbior_kratek& kz = *p;
		
		if (!kz.size())
			continue;

#ifdef DEBUG
		cerr << "grupa: " << (p - zaleznosci.begin()) << ' '; dump_zk(z, kz); cerr << endl;
#endif
		optymalizuj_grupe(kz, do_wpisania);
	}

	set< pair<int, int> >::iterator p = pary_zaleznosci.begin();
	while (p != pary_zaleznosci.end()) {
		set< pair<int, int> >::iterator q = p++;

		if (optymalizuj_miedzy_grupami(zaleznosci[q->first], zaleznosci[q->second], do_wpisania))
			pary_zaleznosci.erase(q);
	}
}

void kombinator::optymalizuj_grupe(zbior_kratek& grupa, zbior_kratek& do_wpisania)
{
	map< int, zbior_kratek > kratki_znaku;
rest:
	for(zbior_kratek::const_iterator p = grupa.begin(); p != grupa.end(); ++p) {
		for(kratka::iterator_mozliwosci q(pole[*p]); q; ++q)
			kratki_znaku[*q].insert(*p);
	}

	int fix = 0;

	for(map< int, zbior_kratek >::iterator p = kratki_znaku.begin(); p != kratki_znaku.end(); /*++p*/) {
		int i = (*p).first;
		zbior_kratek& kz = (*p).second;

		if (kz.size() == 0) {
			kratki_znaku.erase(p++);
			continue;
		} else
			++p;

		if (kz.size() == 1) {
			wpisz_w_kratke(*kz.begin(), i, do_wpisania);
			++fix;
		}
	}
	if (fix) {
#ifdef DEBUG
		cerr << "LV1: restart\n";
#endif
		kratki_znaku.clear();
		goto rest;
	}

	/* tu zostały wszystkie liczby, które występują w co najmniej dwóch kratkach */
	for(map< int, zbior_kratek >::iterator p = kratki_znaku.begin(); p != kratki_znaku.end(); ++p) {
		int i = (*p).first;
		zbior_kratek& kz = (*p).second;

		vector<int> znaki;
		if (optymalizuj_podgrupe(kz, znaki, do_wpisania)) {
			/* usuń inne znaki podgrupy z mapy - żeby ich dwa razy nie oglądać */
			for(vector<int>::const_iterator q = znaki.begin(); q != znaki.end(); ++q) {
				if (*q != i)
					kratki_znaku.erase(*q);
			}
		}
	}
}

bool kombinator::optymalizuj_podgrupe(zbior_kratek& kz, vector<int>& znaki, zbior_kratek& do_wpisania)
{
	/* kz ma co najmniej dwie kratki */
	int nr = *kz.begin();
	const kratka& k = pole[nr];

	/* liczba możliwości == liczba kratek ? */
	if ((unsigned)k.liczba_mozliwosci() != kz.size())
		return false;

	/* sprawdź czy wszystkie kratki mają takie same możliwości */
	for(zbior_kratek::const_iterator q = kz.begin(); ++q != kz.end(); /**/) {
		if (k != pole[*q])
			return false;
	}

	/* znaki podgrupy */
	znaki.reserve(k.liczba_mozliwosci());
	for(kratka::iterator_mozliwosci q(k); q; ++q)
		znaki.push_back(*q);
#ifdef DEBUG
	cerr << "LV2: " << kz.size() << ':'; dump_zk(z, kz); cerr << endl;
	cerr << "LV2: znaki: ";
	for(unsigned di = 0; di < znaki.size(); ++di)
		cerr << z.chars[znaki[di]];
	cerr << endl;

//	dump_full_field(z, pole);
#endif

	/* znajdź wszystkie zależne kratki z grup, do których należą
	 * wszystkie kratki ze sprawdzanej podgrupy */
	zbior_kratek zalezne;
	set<int>& grupy = zkratki[nr];
	for(set<int>::iterator p = grupy.begin(); p != grupy.end(); ++p) {
		zbior_kratek& kgr = zaleznosci[*p];
		bool ok = true;
		for(zbior_kratek::const_iterator q = kz.begin(); ++q != kz.end(); /**/) {
			if (kgr.find(*q) == kgr.end()) {
				ok = false;
				break;
			}
		}

		if (!ok)
			continue;
		zalezne.insert(kgr.begin(), kgr.end());
	}
	/* usuń ze zbioru te, które mamy w sprawdzanej podgrupie */
	for(zbior_kratek::const_iterator q = kz.begin(); q != kz.end(); ++q)
		zalezne.erase(*q);
#ifdef DEBUG
	cerr << "LV2: zalezne:"; dump_zk(z, zalezne); cerr << endl;
#endif

	/* wycinamy... */
	for(zbior_kratek::const_iterator p = zalezne.begin(); p != zalezne.end(); ++p) {
		for(vector<int>::const_iterator q = znaki.begin(); q != znaki.end(); ++q) {
			if (pole[*p].usun_mozliwosc(*q))
				do_wpisania.insert(*p);
		}
	}
#ifdef DEBUG
//	dump_full_field(z, pole);
#endif
	return true;
}

template <typename c1_t, typename c2_t, typename i1_t, typename i2_t, typename i3_t>
void set_parts(c1_t s1, const c1_t s1end, c2_t s2, const c2_t s2end, i1_t i1, i2_t i2, i3_t common)
{
	while(s1 != s1end && s2 != s2end) {
		if (*s1 < *s2)
			*i1++ = *s1++;
		else if (*s2 < *s1)
			*i2++ = *s2++;
		else {
			*common++ = *s1++;
			++s2;
		}
	}

	if (s1 != s1end)
		copy(s1, s1end, i1);
	else if (s2 != s2end)
		copy(s2, s2end, i2);
}

void kombinator::optymalizuj_wspolne(zbior_kratek& g1, zbior_kratek& g2, kratka wspolne, zbior_kratek& do_wpisania)
{
	for(zbior_kratek::const_iterator p = g1.begin(); p != g1.end(); ++p)
		if (!wspolne.usun_mozliwosci(pole[*p])) {
#ifdef DEBUG
			cerr << " X";
#endif
			return;
		}
#ifdef DEBUG
	cerr << ' '; wspolne.dump(cerr, z);
#endif
	for(zbior_kratek::const_iterator p = g2.begin(); p != g2.end(); ++p) {
		if (!pole[*p].usun_mozliwosci(wspolne))
			throw int(0);
		if (pole[*p].liczba_mozliwosci() == 1)
			do_wpisania.insert(*p);
	}
}

bool kombinator::optymalizuj_miedzy_grupami(zbior_kratek& grupa1, zbior_kratek& grupa2, zbior_kratek& do_wpisania)
{
	zbior_kratek g1, g2, wspolne;

	set_parts(grupa1.begin(), grupa1.end(), grupa2.begin(), grupa2.end(),
		insert_iterator<zbior_kratek>(g1, g1.end()),
		insert_iterator<zbior_kratek>(g2, g2.end()),
		insert_iterator<zbior_kratek>(wspolne, wspolne.end()));

#ifdef DEBUG
	cerr << "GX: "; dump_zk(z, g1); cerr << " / "; dump_zk(z, wspolne); cerr << " / "; dump_zk(z, g2); cerr << endl;
#endif

	if (wspolne.size() < 2)
		return true;

	zbior_kratek::const_iterator p = wspolne.begin();
	kratka kwspolne = pole[*p++];
	do {
		kwspolne.dodaj_mozliwosci(pole[*p++]);
	} while (p != wspolne.end());

#ifdef DEBUG
	cerr << "CX: "; kwspolne.dump(cerr, z);
#endif
	optymalizuj_wspolne(g1, g2, kwspolne, do_wpisania);
	optymalizuj_wspolne(g2, g1, kwspolne, do_wpisania);
#ifdef DEBUG
	cerr << endl;
#endif

	return false;
}

bool kombinator::kombinuj()
{
	wygeneruj_zaleznosci();

	zbior_kratek k_odgadniete;

	wpisz_poczatkowe_kratki(k_odgadniete);

	do {
		wpisz_odgadniete(k_odgadniete);

		if (!pozostalo)
			return true;
#ifdef DEBUG
		cerr << "UFF, no to kombinujemy...\n";
		dump_full_field(z, pole);
#endif
		optymalizuj_kratki(k_odgadniete);
#ifdef DEBUG
		dump_full_field(z, pole);
#endif
	} while (k_odgadniete.size());
#ifdef DEBUG
	dump_full_field(z, pole);
#endif
	return false;
}

void kombinator::zapisz_rozwiazanie(rozwiazanie& r)
{
	r.resize(pole.size());
	rozwiazanie::iterator q = r.begin();
	for(vector<kratka>::const_iterator p = pole.begin(); p != pole.end(); ++p, ++q)
		*q = ((*p).rozwiazana()) ? (*p).wartosc() : -(*p).liczba_mozliwosci();
}

bool solve(zadanie& z, rozwiazanie& r)
{
	kombinator solver(z);

	bool b = false;
	try {
		b = solver.kombinuj();
	} catch(int) {
#ifdef DEBUG
		cerr << "broken!\n";
#endif
	}
	solver.zapisz_rozwiazanie(r);

	return b;
}

